from django.db import models
from django.utils import timezone
from tea import timestamp as ts


class TimerMixin(models.Model):
    start_time = models.DateTimeField(
        default=ts.now, null=False, blank=True, editable=True
    )
    end_time = models.DateTimeField(
        null=True, blank=True, default=None, editable=True
    )
    duration = models.BigIntegerField(null=False, blank=False, default=0)

    @property
    def running_time(self) -> str:
        end_time = self.end_time or timezone.now()

        return ts.humanize(int((end_time - self.start_time).total_seconds()))

    def start(self):
        self.start_time = timezone.now()
        self.end_time = None
        self.duration = 0

    def stop(self):
        self.end_time = timezone.now()

    def save(self, *args, **kwargs):
        if self.end_time is not None:
            self.duration = int(
                (self.end_time - self.start_time).total_seconds()
            )
        super().save(*args, **kwargs)

    class Meta:
        app_label = "tea_django"
        abstract = True



import uuid
from typing import Dict, Any

from django.db import models
from django.utils.functional import classproperty
from tea import timestamp as ts
from tea_console.table import RichTableMixin


class UUIDBaseModel(models.Model, RichTableMixin):
    """Base model with UUID as a primary key."""

    id = models.UUIDField(default=uuid.uuid4, primary_key=True)

    @classproperty
    def class_name(cls):
        return cls.__name__

    def column_dict(self) -> Dict[str, Any]:
        d = {}
        for field in self._meta.fields:
            if isinstance(field, models.ForeignKey):
                name = f"{field.name}_id"
            else:
                name = field.name
            d[name] = getattr(self, name)
        return d

    def to_dict(self):
        return self.column_dict()

    @classmethod
    def from_dict(cls, d: dict):
        model_d = {}
        for field in cls._meta.fields:
            if isinstance(field, models.DateTimeField):
                field.auto_now = False
                field.auto_now_add = False
                model_d[field.name] = ts.from_utc_str(d[field.name])
            elif isinstance(field, models.DateField):
                field.auto_now = False
                field.auto_now_add = False
                model_d[field.name] = ts.from_utc_str(d[field.name]).date()
            elif isinstance(field, models.ForeignKey):
                key = f"{field.name}_id"
                model_d[key] = uuid.UUID(d[key])
            else:
                model_d[field.name] = d[field.name]
        return cls(**model_d)

    def __str__(self):
        return f"{self.class_name}({self.id})"

    __repr__ = __str__

    class Meta:
        app_label = "tea_django"
        abstract = True



__all__ = [
    "DjangoTeaError",
    "InvalidConfiguration",
    "DatabaseError",
    "ObjectAlreadyExists",
    "ObjectNotFound",
    "MultipleObjectsFound",
]

from typing import Type, Optional
from tea_console.errors import ConsoleTeaError, InvalidConfiguration


class DjangoTeaError(ConsoleTeaError):
    pass


def query_to_string(query: Optional[dict]) -> str:
    if query is None or len(query) == 0:
        return ""
    else:
        return (
            "("
            + ", ".join(f"{key}={value}" for key, value in query.items())
            + ")"
        )


class DatabaseError(DjangoTeaError):
    pass


class ObjectAlreadyExists(DatabaseError):
    def __init__(self, model: Type, query: Optional[dict] = None):
        self.model = model
        self.query = query

        super().__init__(
            message=(
                f"{model.class_name}{query_to_string(self.query)} "
                f"already exists."
            )
        )


class ObjectNotFound(DatabaseError):
    def __init__(self, model: Type, query: Optional[dict] = None):
        """ObjectNotFound error.

        Args:
            model: Model class.
            query: Query dictionary.
        """
        self.model = model
        self.query = query

        super().__init__(
            message=(
                f"{model.class_name}{query_to_string(self.query)} not found."
            )
        )


class MultipleObjectsFound(DatabaseError):
    def __init__(self, model: Type, query: Optional[dict] = None):
        """MultipleObjectsFound error.

        Args:
            model: Model class.
            query: Query dictionary.
        """
        self.model = model
        self.query = query

        super().__init__(
            message=(
                f"{model.class_name}{query_to_string(self.query)} "
                f"multiple objects found."
            )
        )



__all__ = [
    "ColoredMixin",
    "UniqueSlugMixin",
    "NonUniqueSlugMixin",
    "TimerMixin",
    "TimestampedMixin",
]

from tea_django.models.mixins.colored import ColoredMixin
from tea_django.models.mixins.slug import UniqueSlugMixin, NonUniqueSlugMixin
from tea_django.models.mixins.timer import TimerMixin
from tea_django.models.mixins.timestamped import TimestampedMixin



from rest_framework import serializers
from tea_django.consts import COLOR_RE
from rest_framework.exceptions import ValidationError


def color_validator(value):
    if COLOR_RE.match(value) is None:
        raise ValidationError(f"Invalid color string: {value}")
    return value


class ColoredCreateSerializer(serializers.Serializer):
    color = serializers.CharField(max_length=7, validators=[color_validator])


class ColoredUpdateSerializer(serializers.Serializer):
    color = serializers.CharField(
        max_length=7,
        validators=[color_validator],
        required=False,
        allow_null=True,
    )

    def update(self, instance, validated_data):
        color = validated_data.get("color", None)
        if color is not None:
            instance.color = color



from django.db import models


class TimestampedMixin(models.Model):
    """Mixin for models with created_on and updated_on fields.

    Fields:
        created_on: DateTime field.
        updated_on: DateTime field.
    """

    created_on = models.DateTimeField(
        auto_now_add=True, blank=False, null=False, editable=False
    )
    updated_on = models.DateTimeField(
        auto_now=True, blank=False, null=False, editable=False
    )

    class Meta:
        app_label = "tea_django"
        abstract = True



from typing import Optional

from django.db import models
from django.forms.widgets import TextInput
from django.core.validators import RegexValidator
from django.utils.html import mark_safe, SafeString
from django.utils.translation import gettext_lazy as _

from tea_django.consts import COLOR_RE


color_validator = RegexValidator(
    COLOR_RE, _("Enter a valid color."), "invalid color"
)


class ColorField(models.CharField):
    default_validators = [color_validator]

    def __init__(self, *args, **kwargs):
        # It should always be 7
        kwargs["max_length"] = 7
        kwargs.setdefault("default", "#FFFFFF")
        super().__init__(*args, **kwargs)

    def formfield(self, **kwargs):
        kwargs["widget"] = TextInput(attrs={"type": "color"})
        return super(ColorField, self).formfield(**kwargs)


class ColoredMixin(models.Model):
    color = ColorField(null=False, blank=False)

    def rich(self, s: Optional[str] = None) -> str:
        """Return a rich colored markup for the provided string."""
        return f"[{self.color}]{s or self.color}[/{self.color}]"

    def html(self, s: str) -> SafeString:
        """Return a django safe string in current color."""
        return mark_safe(f'<span style="color: {self.color}">{s}</span>')

    class Meta:
        app_label = "tea_django"
        abstract = True



import logging

from django.http import JsonResponse, Http404
from django.core.exceptions import ObjectDoesNotExist
from rest_framework import status as http_statuses
from rest_framework.exceptions import (
    NotAuthenticated,
    AuthenticationFailed,
    ValidationError,
)

from tea_django import errors


logger = logging.getLogger(__name__)


def format_rest_framework_validation_errors(errors):
    """Format error messages for response.

    Args:
        errors (dict): Dictionary of rest framework validation errors.

    Returns:
        dict: {field: error message}
    """
    if isinstance(errors, list):
        return ". ".join(str(e) for e in errors)
    elif isinstance(errors, dict):
        error = ""
        for key, value in errors.items():
            if key != "non_field_errors":
                error += f"{key}: "
            if isinstance(value, list):
                error += ". ".join(str(v) for v in value)
            else:
                error += str(value)
            error += "\n"
        return error.rstrip("\n")
    else:
        return str(errors)


def tea_django_exception_handler(exc, context):
    if isinstance(exc, (NotAuthenticated, AuthenticationFailed)):
        data = {"error": str(exc)}
        status = http_statuses.HTTP_401_UNAUTHORIZED
    elif isinstance(
        exc, (ObjectDoesNotExist, Http404, errors.ObjectNotFound),
    ):
        data = {"error": str(exc)}
        status = http_statuses.HTTP_404_NOT_FOUND
    elif isinstance(exc, ValidationError):
        data = {"error": format_rest_framework_validation_errors(exc.detail)}
        status = http_statuses.HTTP_400_BAD_REQUEST
    else:
        logger.error("API Error: %s", str(exc))
        data = {"error": str(exc)}
        status = http_statuses.HTTP_500_INTERNAL_SERVER_ERROR
    return JsonResponse(data, status=status)



import random
import string

from slugify import slugify
from django.db import models
from django.utils.crypto import get_random_string


def generate_random_slug():
    return get_random_string(8, "0123456789")  # 8 characters, only digits.


def create_unique_slug(value, model, object_pk, max_length=45):
    """Create a unique slug from value.

    Args:
        value (str): Value to slugify.
        model: Model to check against for uniqueness.
        object_pk: Object PK of the one we are creating the slug for.
        max_length (int): Maximal length of the slug before appending the
            number to provide uniqueness.

    Returns:
        str: Unique slug.
    """
    # Always exclude the existing object this is important because when we set
    # slug=None, while this values is not yet saved to the database, it will
    # come up as a collision with the saved database value of the same object.
    queryset = model.objects.exclude(pk=object_pk)

    # First create a slug
    slug_words = slugify(value).split("-")
    slug = slug_words[0]
    for word in slug_words[1:]:
        new_slug = f"{slug}-{word}"
        if len(new_slug) <= max_length:
            slug = new_slug
        else:
            break

    # Now check for uniqueness
    slug_count = queryset.filter(slug=slug).count()
    if slug_count == 0:
        return slug
    else:
        for i in range(slug_count, 10000):
            new_slug = f"{slug}-{i}"
            if queryset.filter(slug=new_slug).count() == 0:
                return new_slug

        # This should never happen but let's be prepared
        word = "".join(random.choice(string.ascii_lowercase) for _ in range(4))
        return f"{slug}-{word}"


class SlugMixinBase(models.Model):
    """Base class for unique and non unique slugs."""

    SLUG_FIELD = "name"

    def _slug_value(self):
        """Return a string value from which the slug will be created."""
        return getattr(self, self.SLUG_FIELD, str(self))

    @classmethod
    def get_by_slug(cls, slug: str, **kwargs):
        """Get an object by it's slug.

        Args:
             slug (str): Search slug

        Returns:
            Queried object.

        Raises:
            ObjectNotFound: If the object is not found.
        """
        return cls.objects.get(slug=slug, **kwargs)

    @classmethod
    def get_by_slug_field(cls, value: str, **kwargs):
        return cls.get_by_slug(slug=slugify(value), **kwargs)

    class Meta:
        app_label = "tea_django"
        abstract = True


class UniqueSlugMixin(SlugMixinBase):
    """Mixin for models with unique slug field.

    Fields:
        slug: Slug field.
    """

    slug = models.SlugField(unique=True)

    def save(self, *args, **kwargs):
        # FIXME: Do this only when the slug field changes.
        self.slug = create_unique_slug(
            value=self._slug_value(), model=self.__class__, object_pk=self.pk,
        )
        super().save(*args, **kwargs)

    class Meta:
        app_label = "tea_django"
        abstract = True


class NonUniqueSlugMixin(SlugMixinBase):
    """Mixin for models with non-unique slug field.

    Fields:
        slug: Slug field.
    """

    slug = models.SlugField()

    def save(self, *args, **kwargs):
        self.slug = slugify(self._slug_value())
        super().save(*args, **kwargs)

    class Meta:
        app_label = "tea_django"
        abstract = True



import base64
from typing import Dict
from random import SystemRandom

from tea_console.config import ConfigField, Config as ConsoleConfig

from tea_django import consts


def to_value(v: str) -> str:
    return base64.b64decode(v.encode("utf-8")).decode("utf-8")


def to_string(v: str) -> str:
    return base64.b64encode(v.encode("utf-8")).decode("utf-8")


class Config(ConsoleConfig):

    ENTRIES: Dict[str, ConfigField] = {
        **ConsoleConfig.ENTRIES,
        "secret_key": ConfigField(
            section="django",
            option="secret_key",
            to_value=to_value,
            to_string=to_string,
        ),
    }

    def __init__(self, config_file):
        self.secret_key = "".join(
            SystemRandom().choice(consts.SECRETE_KEY_ALLOWED_CHARS)
            for _ in range(50)
        )
        super().__init__(config_file=config_file)



class Version:
    __slots__ = ("major", "minor", "build")

    def __init__(self, major, minor, build):
        self.major = major
        self.minor = minor
        self.build = build

    def __str__(self):
        return f"{self.major}.{self.minor}.{self.build}"

    def __repr__(self):
        return (
            f"Version(major={self.major}, minor={self.minor}, "
            f"build={self.build})"
        )


version = Version(0, 0, 6)
__version__ = str(version)



import io
import importlib.util
from pathlib import Path
from setuptools import setup, find_packages

author = "Viktor Kerkez"
author_email = "alefnula@gmail.com"


def get_version():
    """Import the version module and get the project version from it."""
    version_py = Path(__file__).parent / "tea_django" / "version.py"
    spec = importlib.util.spec_from_file_location("version", version_py)
    version = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(version)
    return version.__version__


setup(
    name="tea-django",
    version=get_version(),
    author=author,
    author_email=author_email,
    maintainer=author,
    maintainer_email=author_email,
    description="Set of django helpers for developing web and non-web apps.",
    long_description=io.open("README.md", "r", encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/alefnula/tea-django",
    platforms=["Windows", "POSIX", "MacOSX"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    license="Apache-2.0",
    packages=find_packages(),
    include_package_data=True,
    install_requires=io.open("requirements.txt").read().splitlines(),
)



from tea_console.table import RichTableMixin
from django.utils.functional import classproperty


class VanillaModel(RichTableMixin):
    @classproperty
    def class_name(cls):
        return cls.__name__

    def to_dict(self) -> dict:
        return {}
